from __future__ import annotations

from glob import iglob
from os import path
from typing import Final

from more_itertools import first

from hyprshade.utils import hypr_config_home

from . import hyprctl


def _stripped_basename(s: str) -> str:
    return path.splitext(path.basename(s))[0]


class _ShaderDirs:
    ENV_VAR_NAME: Final = "HYPRSHADE_SHADERS_DIR"

    @staticmethod
    def env() -> str:
        return path.expanduser(
            path.expandvars(path.expandvars("$" + _ShaderDirs.ENV_VAR_NAME))
        )

    SYSTEM_DIR: Final = "/usr/share/hyprshade/shaders"

    @staticmethod
    def system() -> str:
        return _ShaderDirs.SYSTEM_DIR

    @staticmethod
    def user() -> str:
        return path.join(hypr_config_home(), "shaders")

    @staticmethod
    def all() -> list[str]:
        return [
            x
            for x in [_ShaderDirs.env(), _ShaderDirs.system(), _ShaderDirs.user()]
            if path.exists(x)
        ]


class Shader:
    dirs: Final = _ShaderDirs

    def __init__(self, shader_name_or_path: str):
        self._given_path = (
            shader_name_or_path
            if path.dirname(shader_name_or_path) and path.isfile(shader_name_or_path)
            else None
        )
        self._name = _stripped_basename(shader_name_or_path)

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, Shader):
            return False
        s, s2 = self._resolve_path(), __value._resolve_path()
        return path.samefile(s, s2)

    def __str__(self) -> str:
        return self._name

    def __repr__(self) -> str:
        return f"Shader({self._name!r})"

    @property
    def name(self) -> str:
        return self._name

    def on(self) -> None:
        path_ = self._resolve_path()
        hyprctl.set_screen_shader(path_)

    @staticmethod
    def off() -> None:
        hyprctl.clear_screen_shader()

    @staticmethod
    def current() -> Shader | None:
        name = hyprctl.get_screen_shader()
        return None if name is None else Shader(name)

    def _resolve_path(self) -> str:
        if self._given_path:
            return self._given_path

        dirs = Shader.dirs.all()
        for dir in dirs:
            path_ = first(iglob(f"{self._name}*", root_dir=dir), None)
            if path_ is not None:
                return path.join(dir, path_)

        raise FileNotFoundError(
            f"Shader '{self._name}' could not be found in any of the following"
            " directories:\n\t"
            "{}".format("\n\t".join(dirs))
        )



from pathlib import Path

import pytest
from hyprshade import hyprctl


@pytest.mark.hyprland()
@pytest.mark.usefixtures("_save_screen_shader")
class TestHyprctl:
    def test_hyprctl(self, shader_path: Path):
        hyprctl.set_screen_shader(str(shader_path))
        assert hyprctl.get_screen_shader() == str(shader_path)

    def test_clear(self):
        hyprctl.clear_screen_shader()
        assert hyprctl.get_screen_shader() is None



# mypy: disable-error-code="arg-type"

import click
import pytest
from hyprshade import click_utils


class TestConvertToShader:
    def test_works(self):
        shader = click_utils.convert_to_shader(None, None, "foo")
        assert shader.name == "foo"

    def test_none(self):
        assert click_utils.convert_to_shader(None, None, None) is None


class TestValidateOptionalParam:
    def test_empty(self):
        assert click_utils.validate_optional_param(None, None, ()) is None

    def test_single(self):
        assert click_utils.validate_optional_param(None, None, ("foo",)) == "foo"

    def test_multiple(self):
        with pytest.raises(click.UsageError):
            click_utils.validate_optional_param(None, None, ("foo", "bar"))


class TestOptionalParam:
    def test_basic(self):
        o = click_utils.optional_param("FOO", None)
        assert o["metavar"] == "FOO"
        assert o["nargs"] == -1
        assert o["callback"] is not None

    def test_callback(self):
        o = click_utils.optional_param("FOO", None)
        callback = o["callback"]
        assert callback(None, None, ("foo",)) == "foo"
        assert callback(None, None, ()) is None

    def test_given_callback(self):
        o = click_utils.optional_param("FOO", lambda *_: "bar")
        callback = o["callback"]
        assert callback(None, None, ("foo",)) == "bar"

    def test_callback_synergy(self):
        o = click_utils.optional_param("FOO", lambda _1, _2, x: x + "bar")
        callback = o["callback"]
        assert callback(None, None, ("foo",)) == "foobar"



from pathlib import Path

import pytest
from hyprshade import hyprctl


@pytest.fixture(scope="module")
def _save_screen_shader():
    screen_shader = hyprctl.get_screen_shader()
    yield

    if screen_shader is None:
        hyprctl.clear_screen_shader()
        return
    try:
        hyprctl.set_screen_shader(screen_shader)
    except BaseException:
        import os

        hyprctl.clear_screen_shader()
        os.system('notify-send "hyprshade" "Failed to restore screen shader"')


@pytest.fixture()
def shader_path(tmp_path: Path):
    shader_path = tmp_path / "shader.frag"
    shader_path.write_text("void main() {}")
    return shader_path



from __future__ import annotations

from os import path
from typing import Literal

from hyprshade.utils import systemd_user_config_home

SystemdUnitType = Literal["service", "timer"]


def write_systemd_user_unit(unit_type: SystemdUnitType, body: str):
    with open(
        path.join(systemd_user_config_home(), f"hyprshade.{unit_type}"), "w"
    ) as f:
        f.write(body)



from pathlib import Path

import pytest
from hyprshade import helpers


def test_write_systemd_user_unit(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
    (tmp_path / "systemd" / "user").mkdir(parents=True)

    helpers.write_systemd_user_unit("service", "foo")
    assert (tmp_path / "systemd" / "user" / "hyprshade.service").read_text() == "foo"

    helpers.write_systemd_user_unit("timer", "bar")
    assert (tmp_path / "systemd" / "user" / "hyprshade.timer").read_text() == "bar"



from pathlib import Path

import pytest
from hyprshade.shader import Shader, _stripped_basename


class TestDirs:
    def test_env(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setenv(Shader.dirs.ENV_VAR_NAME, "/foo/bar")
        assert Shader.dirs.env() == "/foo/bar"
        monkeypatch.delenv(Shader.dirs.ENV_VAR_NAME, raising=False)
        assert Shader.dirs.env() == "$" + Shader.dirs.ENV_VAR_NAME

    def test_system(self):
        assert Shader.dirs.system() == Shader.dirs.SYSTEM_DIR

    def test_user(self, monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
        monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
        assert Shader.dirs.user() == str(tmp_path / "hypr/shaders")

    def test_all(self, monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
        env_path = tmp_path / "hypr/env/shaders"
        env_path.mkdir(parents=True)
        monkeypatch.setenv(Shader.dirs.ENV_VAR_NAME, str(env_path))

        system_path = tmp_path / "hypr/system/shaders"
        system_path.mkdir(parents=True)
        Shader.dirs.SYSTEM_DIR = str(system_path)  # type: ignore[misc]

        user_path = tmp_path / "hypr/shaders"
        user_path.mkdir(parents=True)
        monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))

        assert Shader.dirs.all() == list(map(str, [env_path, system_path, user_path]))


class TestConstructor:
    def test_name(self):
        assert Shader("foo")._name == "foo"

    def test_path(self, shader_path: Path):
        shader = Shader(str(shader_path))
        assert shader._name == "shader"
        assert shader._given_path == str(shader_path)

    def test_nonexistent_path(self, tmp_path: Path):
        shader_path = tmp_path / "foo.glsl"
        assert not shader_path.exists(), "test assumption failed"

        shader = Shader(str(shader_path))
        assert shader._name == "foo"
        assert shader._given_path is None


class TestEquality:
    def test_same_path(self, monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
        monkeypatch.setenv(Shader.dirs.ENV_VAR_NAME, str(tmp_path))
        shader_path = tmp_path / "foo.glsl"
        shader_path.touch()

        assert Shader("foo") == Shader("foo")
        assert Shader(str(shader_path)) == Shader(str(shader_path))
        assert Shader("foo") == Shader(str(shader_path))
        assert Shader(str(shader_path)) == Shader("foo")

    def test_different_path(self, monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
        monkeypatch.setenv(Shader.dirs.ENV_VAR_NAME, str(tmp_path))
        shader_path1 = tmp_path / "foo.glsl"
        shader_path2 = tmp_path / "bar.glsl"
        shader_path1.touch()
        shader_path2.touch()

        assert Shader("foo") != Shader("bar")
        assert Shader(str(shader_path1)) != Shader(str(shader_path2))
        assert Shader("foo") != Shader(str(shader_path2))
        assert Shader(str(shader_path1)) != Shader("bar")

    def test_other(self):
        assert Shader("foo") != None  # noqa: E711
        assert Shader("foo") != "foo"


class TestDisplay:
    def test_str(self):
        assert str(Shader("foo")) == "foo"

    def test_repr(self):
        assert repr(Shader("foo")) == "Shader('foo')"

    def test_name(self):
        assert Shader("foo").name == "foo"


@pytest.mark.hyprland()
@pytest.mark.usefixtures("_save_screen_shader")
class TestOnOff:
    def test_on(self, monkeypatch: pytest.MonkeyPatch, shader_path: Path):
        monkeypatch.setenv(Shader.dirs.ENV_VAR_NAME, str(shader_path.parent))
        name = _stripped_basename(str(shader_path))
        Shader(name).on()
        assert Shader.current() == Shader(name)

    def test_on_path(self, shader_path: Path):
        Shader(str(shader_path)).on()
        assert Shader.current() == Shader(str(shader_path))

    def test_on_doesnotexist(self):
        with pytest.raises(FileNotFoundError):
            Shader("doesnotexist").on()

    def test_off(self):
        Shader.off()
        assert Shader.current() is None


class TestResolvePath:
    def test_one_path(self, monkeypatch: pytest.MonkeyPatch, shader_path: Path):
        monkeypatch.setenv(Shader.dirs.ENV_VAR_NAME, str(shader_path.parent))
        assert Shader("shader")._resolve_path() == str(shader_path)

    def test_env_priority(
        self, monkeypatch: pytest.MonkeyPatch, shader_path: Path, tmp_path: Path
    ):
        system_path = tmp_path / "hypr/system/shaders"
        system_path.mkdir(parents=True)
        (system_path / "shader.glsl").touch()
        assert (system_path / "shader.glsl").exists(), "test assumption failed"

        Shader.dirs.SYSTEM_DIR = str(tmp_path)  # type: ignore[misc]
        monkeypatch.setenv(Shader.dirs.ENV_VAR_NAME, str(shader_path.parent))

        assert Shader("shader")._resolve_path() == str(shader_path)

    def test_given_path(self, shader_path: Path):
        assert Shader(str(shader_path))._resolve_path() == str(shader_path)

    def test_notfound(self, tmp_path: Path):
        with pytest.raises(FileNotFoundError):
            Shader("foo")._resolve_path()

        path_not_exist = tmp_path / "foo.glsl"
        assert not path_not_exist.exists(), "test assumption failed"
        with pytest.raises(FileNotFoundError):
            Shader(str(path_not_exist))._resolve_path()

    def test_ignores_cwd(self, monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
        shader_path = tmp_path / "foo.glsl"
        shader_path.touch()
        monkeypatch.chdir(tmp_path)
        with pytest.raises(FileNotFoundError):
            Shader("foo")._resolve_path()



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

import click

from .shader import Shader

if TYPE_CHECKING:
    from collections.abc import Callable


def convert_to_shader(
    ctx: click.Context, param: click.Argument, shader_name_or_path: str | None
):
    return Shader(shader_name_or_path) if shader_name_or_path is not None else None


T = TypeVar("T", str, int, float, bool, click.ParamType)


def validate_optional_param(
    ctx: click.Context, param: click.Argument, value: tuple[T, ...]
) -> T | None:
    if extra_args := value[1:]:
        s = "s" if len(extra_args) > 1 else ""
        raise click.UsageError(
            f"Got unexpected extra argument{s} ({' '.join(map(str, extra_args))})"
        )

    return None if len(value) == 0 else value[0]


def optional_param(metavar: str | None = None, callback: Callable | None = None):
    def merged_callback(
        ctx: click.Context, param: click.Argument, value: tuple[T, ...]
    ):
        value2 = validate_optional_param(ctx, param, value)
        if callback is not None:
            return callback(ctx, param, value2)
        return value2

    return {
        "metavar": metavar,
        "nargs": -1,
        "callback": merged_callback,
    }



from __future__ import annotations

import logging
import sys
from datetime import datetime, time

import click
from more_itertools import quantify

from .click_utils import convert_to_shader, optional_param
from .config import Config
from .helpers import write_systemd_user_unit
from .shader import Shader
from .utils import ls_dirs


@click.group()
@click.version_option()
@click.option("-v", "--verbose", is_flag=True, help="Enable verbose output.")
def cli(verbose: bool):
    level = logging.DEBUG if verbose else logging.WARNING
    logging.basicConfig(level=level)


def main():
    try:
        cli()
    except Exception as e:
        if logging.getLogger().getEffectiveLevel() <= logging.DEBUG:
            raise e
        click.echo(f"Error: {e}", err=True)
        sys.exit(1)


@cli.command()
@click.argument("shader", callback=convert_to_shader)
def on(shader: Shader):
    """Turn on screen shader."""

    shader.on()


@cli.command()
def off():
    """Turn off screen shader."""

    Shader.off()


def get_shader_to_toggle(
    shader: Shader | None, fallback: Shader | None
) -> Shader | None:
    current = Shader.current()
    if shader == current:
        return fallback
    return shader


def get_fallback(
    *,
    shader: Shader | None,
    default: Shader | None,
    auto: Shader | None,
    fallback_default: bool,
    fallback_auto: bool,
) -> Shader | None:
    if fallback_default or (fallback_auto and shader == auto):
        return default
    elif fallback_auto:
        return auto
    return None


def try_from_config(t: time, panic: bool) -> tuple[Shader | None, Shader | None]:
    try:
        schedule = Config().to_schedule()
    except FileNotFoundError:
        if panic:
            raise
        return None, None
    return schedule.scheduled_shader(t), schedule.default_shader


@cli.command()
@click.argument("shader", **optional_param("SHADER", convert_to_shader))
@click.option(
    "--fallback",
    metavar="SHADER",
    callback=convert_to_shader,
    help="Shader to switch to instead of toggling off.",
)
@click.option(
    "--fallback-default",
    is_flag=True,
    default=False,
    help="Use default shader as fallback.",
)
@click.option(
    "--fallback-auto",
    is_flag=True,
    default=False,
    help="Use currently scheduled shader as fallback. Equivalent to --fallback-default"
    " when the currently scheduled shader is SHADER.",
)
def toggle(
    shader: Shader | None,
    fallback: Shader | None,
    fallback_default: bool,
    fallback_auto: bool,
):
    """Toggle screen shader.

    If run with no arguments, SHADER is inferred based on schedule.

    When a fallback shader is specified, will instead toggle between SHADER
    and the fallback shader.
    """

    t = datetime.now().time()

    fallback_opts = [fallback, fallback_default, fallback_auto]
    if quantify(fallback_opts) > 1:
        raise click.BadOptionUsage(
            "--fallback", "Cannot specify more than 1 --fallback* option"
        )

    scheduled, default = try_from_config(t, fallback_default or fallback_auto)
    shader = shader or scheduled

    fallback = fallback or get_fallback(
        shader=shader,
        default=default,
        auto=scheduled,
        fallback_default=fallback_default,
        fallback_auto=fallback_auto,
    )
    shader_to_toggle = get_shader_to_toggle(shader, fallback)
    if shader_to_toggle:
        shader_to_toggle.on()
    else:
        Shader.off()


@cli.command()
@click.pass_context
def auto(ctx: click.Context):
    """Turn on/off screen shader based on schedule."""

    t = datetime.now().time()
    shader = Config().to_schedule().scheduled_shader(t)

    if shader:
        shader.on()
    else:
        Shader.off()


@cli.command()
def install():
    """Install systemd user units."""

    schedule = Config().to_schedule()
    timer_config = "\n".join(
        sorted([f"OnCalendar=*-*-* {x}" for x in schedule.event_times()])
    )

    write_systemd_user_unit(
        "service",
        """[Unit]
Description=Apply screen filter

[Service]
Type=oneshot
ExecStart="/usr/bin/hyprshade" auto
""",
    )

    write_systemd_user_unit(
        "timer",
        f"""[Unit]
Description=Apply screen filter on schedule

[Timer]
{timer_config}

[Install]
WantedBy=timers.target
""",
    )


@cli.command()
def ls():
    """List available screen shaders."""

    current = Shader.current()

    for shader in map(Shader, ls_dirs(Shader.dirs.all())):
        c = "*" if shader == current else " "
        click.echo(f"{c} {shader}")



from __future__ import annotations

import os
from os import PathLike, path
from typing import TYPE_CHECKING

from more_itertools import flatten, unique_justseen

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from datetime import time


def xdg_config_home():
    config_home = os.getenv("XDG_CONFIG_HOME")
    if config_home is None:
        home = os.getenv("HOME")
        if home is None:
            raise ValueError("$HOME environment variable is not set")
        config_home = path.join(home, ".config")
    return config_home


def hypr_config_home():
    return path.join(xdg_config_home(), "hypr")


def hyprshade_config_home():
    return path.join(xdg_config_home(), "hyprshade")


def systemd_user_config_home():
    return path.join(xdg_config_home(), "systemd/user")


def is_time_between(time_: time, start_time: time, end_time: time) -> bool:
    if end_time <= start_time:
        return start_time <= time_ or time_ <= end_time
    return start_time <= time_ <= end_time


def ls_dirs(dirs: Iterable[str | PathLike[str]]) -> Iterator[str]:
    return unique_justseen(sorted(flatten(map(os.listdir, dirs))))



from __future__ import annotations

import json
import subprocess
from json import JSONDecodeError
from os import path
from typing import Final

EMPTY_STR: Final = "[[EMPTY]]"


def set_screen_shader(shader_path: str) -> None:
    subprocess.run(
        ["hyprctl", "keyword", "decoration:screen_shader", shader_path],
        capture_output=True,
        check=True,
    )


def clear_screen_shader() -> None:
    set_screen_shader(EMPTY_STR)


def get_screen_shader() -> str | None:
    try:
        hyprctl_pipe = subprocess.run(
            ["hyprctl", "-j", "getoption", "decoration:screen_shader"],
            capture_output=True,
            check=True,
            encoding="utf-8",
        )
        # TODO: Remove sed workaround
        # (when hyprwm/Hyprland@4743041 is pushed in a stable release)
        hyprctl_pipe = subprocess.run(
            ["sed", "/^adding/d"],
            input=hyprctl_pipe.stdout,
            capture_output=True,
            check=True,
            encoding="utf-8",
        )
        shader_json = json.loads(hyprctl_pipe.stdout)
    except JSONDecodeError as e:
        raise RuntimeError("Failed to parse JSON returned by hyprctl") from e

    shader = str(shader_json["str"]).strip()
    if shader == EMPTY_STR:
        return None
    if not path.isfile(shader):
        raise RuntimeError(f"Got shader {shader} from hyprctl, which does not exist")

    return shader



from __future__ import annotations

import os
import tomllib
from datetime import time
from itertools import chain, pairwise
from os import path
from typing import TYPE_CHECKING, Literal, NotRequired, TypedDict, cast

from more_itertools import first_true, nth, partition

from .shader import Shader
from .utils import hypr_config_home, hyprshade_config_home, is_time_between

if TYPE_CHECKING:
    from collections.abc import Iterator

TimeRange = tuple[time, time]


class ShaderConfig(TypedDict):
    name: str
    start_time: time
    end_time: NotRequired[time]


class DefaultShadeConfig(TypedDict):
    name: str
    default: Literal[True]


class ConfigDict(TypedDict):
    shades: list[ShaderConfig | DefaultShadeConfig]


class ScheduleEntry:
    shader: Shader
    start_time: time
    end_time: time | None

    def __init__(self, shader_config: ShaderConfig):
        self.shader = Shader(shader_config["name"])
        self.start_time = shader_config["start_time"]
        self.end_time = shader_config.get("end_time")

    def __repr__(self) -> str:
        return (
            f'ScheduleEntry("{self.shader}", start_time={self.start_time}, '
            f"end_time={self.end_time})"
        )


class Schedule:
    entries: list[ScheduleEntry]
    default_shader: Shader | None

    def __init__(self, config: Config):
        rest_shader_configs, default_shader_config = config.partition()
        sorted_rest_shader_configs = sorted(
            rest_shader_configs, key=lambda s: s["start_time"]
        )
        self.entries = list(map(ScheduleEntry, sorted_rest_shader_configs))
        self.default_shader = (
            Shader(default_shader_config["name"])
            if default_shader_config is not None
            else None
        )

    def scheduled_shader(self, t: time) -> Shader | None:
        for shader, (start_time, end_time) in self._resolved_entries():
            if is_time_between(t, start_time, end_time):
                return shader

        return self.default_shader

    def event_times(self) -> Iterator[time]:
        for entry in self.entries:
            yield entry.start_time
            if entry.end_time is not None:
                yield entry.end_time

    def _resolved_entries(self) -> Iterator[tuple[Shader, TimeRange]]:
        for entry, next_entry in pairwise(chain(self.entries, [self.entries[0]])):
            start_time = entry.start_time
            end_time = entry.end_time or next_entry.start_time
            yield entry.shader, (start_time, end_time)


class Config:
    _dict: ConfigDict

    def __init__(self, path_: str | None = None):
        path_ = path_ or Config.get_path()
        if path_ is None:
            raise FileNotFoundError("Config file not found")
        self._dict = Config._load(path_)

    @staticmethod
    def _load(path_: str) -> ConfigDict:
        with open(path_, "rb") as f:
            return cast(ConfigDict, tomllib.load(f))

    @staticmethod
    def get_path() -> str | None:
        candidates = [
            os.getenv("HYPRSHADE_CONFIG"),
            path.join(hypr_config_home(), "hyprshade.toml"),
            path.join(hyprshade_config_home(), "config.toml"),
        ]
        return first_true((c for c in candidates if c is not None), pred=path.isfile)

    def partition(self) -> tuple[list[ShaderConfig], DefaultShadeConfig | None]:
        no_default, yes_default = partition(
            lambda s: s.get("default", False), self._dict["shades"]
        )
        rest = cast(list[ShaderConfig], list(no_default))
        default = cast(DefaultShadeConfig, nth(yes_default, 0))

        # TODO: Should raise an exception instead of asserting
        assert nth(yes_default, 0) is None, "There should be only one default shade"

        return rest, default

    def to_schedule(self) -> Schedule:
        return Schedule(self)
